#region - Terms of Usage -
/*
 * Copyright 2006 Sameera Perera
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
*/
#endregion

using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.ComponentModel;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;

namespace Inferno.Utilities
{
    #region : The Marker :
    public class Marker
    {


        public event EventHandler MarkerUpdated;

        private RectangleF bounds;
        public RectangleF Bounds
        {
            get { return bounds; }
            set 
            {
                if ( this.IsMovable )
                    bounds = value;
            }
        }

        private Color color;
        public Color Color
        {
            get { return color; }
            set
            {
                color = value;
                if ( MarkerUpdated != null )
                    MarkerUpdated( this, EventArgs.Empty );
            }
        }

        private float position;
        public float Position
        {
            get { return position; }
            set
            {
                position = value;
                if ( MarkerUpdated != null )
                    MarkerUpdated( this, EventArgs.Empty );
            }
        }

        public readonly bool IsMovable;

        private const int MARKER_SIZE     = 10;

        // Set the Position value without causing the MarkerUpdated event to fire
        internal void SetPositionSilently( float position )
        {
            this.position = position;
        }

        // Update the marker's bounds based on the marker strip's bounds.
        public void UpdateBounds( Rectangle markerStripBounds )
        {
            float xOffset    = markerStripBounds.Width * Position + markerStripBounds.X;
            float adjustment = MARKER_SIZE / 2;

            if ( Position < 1.0 ) xOffset -= adjustment;
            else xOffset += adjustment;

            this.bounds = new RectangleF(
                                new PointF( xOffset, markerStripBounds.Top ),
                                new SizeF( MARKER_SIZE, MARKER_SIZE ) );
        }

        public Marker(float position, Color color) : this(position, color, true) {}

        public Marker(float position) : this(position, Color.White, true) {}

        public Marker( float position, Color color, bool isMovable )
        {
            this.Position   = position;
            this.Color      = color;
            this.IsMovable  = isMovable;
        }
    }
    #endregion
    public class MarkerList
    {

        public enum RFunctions { z, sin,cos,sqr, log,logPlus1, sqrt, tang, pow};

        private List<Marker> mMarkers;
        RFunctions mConvertFunction;
        double mPower;
        double mLogBase;

        bool    Changes;

        public RFunctions Function
        {
            get { return mConvertFunction; }
            set { Changes = true;  mConvertFunction = value; }
        }
        public double Power
        {
            get { return mPower; }
            set { Changes = true;  mPower = value; }
        }
        public double LogBase
        {
            get { return mLogBase; }
            set { Changes = true; mLogBase = value; }
        }


        public MarkerList() 
        { 
                mMarkers = new List<Marker>();
                mMarkers.Add(new Marker(0, Color.Black));
                mMarkers.Add(new Marker((float)0.1, Color.Blue));
                mMarkers.Add(new Marker((float)0.20, Color.Red));
                mMarkers.Add(new Marker((float)0.40, Color.Aquamarine));
                mMarkers.Add(new Marker((float)0.60, Color.Green));
                mMarkers.Add(new Marker((float)0.80, Color.CornflowerBlue));
                mMarkers.Add(new Marker((float)1, Color.White));

                mConvertFunction = RFunctions.z;
                mPower = 0.3;
                LogBase = 1.5;
                Changes = true;

        }
        public bool CheckChange()
        {
            bool result = Changes;
            Changes = false;

            return result;
        }
        public void Change()
        {
            Changes = true;
                
        }
        public double ConvertFunction(double iValue){

            switch (mConvertFunction)
            { 
                case RFunctions.z:
                            return iValue;
                case RFunctions.sin:
                    return Math.Abs(Math.Sin(iValue));
                case RFunctions.cos:
                    return Math.Abs(Math.Cos(iValue));
                case RFunctions.sqr:
                    return iValue * iValue;
                case RFunctions.log:
                    return Math.Log(iValue, LogBase);
                case RFunctions.logPlus1:
                    return Math.Log(iValue + 1, LogBase);
                case RFunctions.sqrt:
                    return Math.Sqrt(iValue);
                case RFunctions.tang:
                    return Math.Tan(iValue);
                case RFunctions.pow:
                    return Math.Pow(iValue, mPower);
            }
            return iValue;
        }
        /*
        public List<Marker> markers
        {
            get
            {
                return mMarkers;
            }
            set
            {
                mMarkers = value;            
            }
        }
         */

        public List<Marker> GetMarkers() { return mMarkers; }
        public void  SetMarkers(List<Marker> nMarkers) {mMarkers = nMarkers; }
        public Color GetColor(double position) 
        {
            double posKoef;
            Color  Result= Color.Black;
            double R, G, B, A;
            
            for (int i = 1; i < mMarkers.Count; i++) 
            {
                if (( (mMarkers[i]).Position >= position) &&(mMarkers[i-1]).Position <= position)
                {
                    
                    posKoef = (position -  mMarkers[i-1].Position ) / (mMarkers[i].Position - mMarkers[i - 1].Position);
                   R = mMarkers[i-1].Color.R +
                              posKoef * (mMarkers[i].Color.R - mMarkers[i - 1].Color.R);
                   G = mMarkers[i-1].Color.G +
                              posKoef * (mMarkers[i].Color.G - mMarkers[i - 1].Color.G);
                   B = mMarkers[i-1].Color.B +
                              posKoef * (mMarkers[i].Color.B - mMarkers[i - 1].Color.B);
                   A = mMarkers[i-1].Color.A +
                              posKoef * (mMarkers[i].Color.A - mMarkers[i - 1].Color.A);


                   Result = Color.FromArgb((int)A,(int)R, (int)G, (int)B);
                } 
            }

            return Result;
        }
    }

    class GradientBuilder : Control
    {
        #region Memeber Variables

        private Rectangle gradBarRect;
        private Rectangle markerStrip;
        private Rectangle compansatedMarkerStrip;

        public  MarkerList mMarkers;

        /*private List<Marker> markers 
        {
            get 
            {
                return mMarkers.markers;
            }
            set 
            {
                mMarkers.markers = value;
            
            }
        }
        */

        public List<Marker> GetMarkers() { return mMarkers.GetMarkers(); }
        public void SetMarkers(List<Marker> nMarker) { mMarkers.SetMarkers(nMarker);}
        private const int DEFAULT_PADDING = 10;

        private Marker selectedMarker   = null;
        private Marker movingMarker     = null;
        private Bitmap gradientBmp      = null;
        private bool isBitmapInvalid    = true;

        #endregion

        #region Public Memembers

        public event EventHandler MarkerSelected;
        public event EventHandler GradientChanged;

        private Gradient gradient;
        [DefaultValue(null), Browsable(false)]
        public Gradient Gradient
        {
            get { return gradient; }
            set 
            { 
                gradient = value; 
                // Create markers from the gradient
                ColorBlend blend    = gradient.ColorBlend;

                GetMarkers().Clear();

                Marker marker = new Marker( 0.0f, blend.Colors[ 0 ], false );
                marker.MarkerUpdated += new EventHandler( marker_MarkerUpdated );
                GetMarkers().Add(marker);

                for ( int i = 1; i < blend.Colors.Length - 1; i++ )
                {
                    marker = new Marker( blend.Positions[ i ], blend.Colors[ i ] );
                    marker.MarkerUpdated += new EventHandler( marker_MarkerUpdated );
                    GetMarkers().Add(marker);
                }

                marker = new Marker( 1.0f, blend.Colors[ blend.Colors.Length-1 ], false );
                marker.MarkerUpdated += new EventHandler( marker_MarkerUpdated );
                GetMarkers().Add(marker);

                this.Invalidate();
                OnGradientChanged(false);
            }
        }

        private bool silentMarkers = false;
        /// <summary>
        /// Gets or sets whether the GradientChanged event is fired when a marker
        /// is dragged.
        /// </summary>
        [DefaultValue(false)]
        public bool SilentMarkers
        {
            get { return silentMarkers; }
            set { silentMarkers = value; }
        }


        #endregion

        #region Calculating

        private void Recalculate()
        {

            gradBarRect = new Rectangle( DEFAULT_PADDING, DEFAULT_PADDING,
                            this.ClientRectangle.Width - DEFAULT_PADDING * 2,
                            this.ClientRectangle.Height - DEFAULT_PADDING * 3 );
            markerStrip = new Rectangle( DEFAULT_PADDING, gradBarRect.Bottom + 1,
                            gradBarRect.Width,
                            DEFAULT_PADDING + 2 );

            compansatedMarkerStrip   = new Rectangle(
                        markerStrip.X, markerStrip.Y, 
                        markerStrip.Width - DEFAULT_PADDING,
                        markerStrip.Height);
            foreach (Marker mrk in mMarkers.GetMarkers())
                mrk.UpdateBounds( compansatedMarkerStrip );
        }

        protected override void OnResize( EventArgs e )
        {
            base.OnResize( e );
            Recalculate();
        }

        #endregion

        #region Painting

        protected virtual void OnGradientChanged(bool suppressEvent)
        {
            isBitmapInvalid           = true;
            gradient.ColorBlend     = GetBlendFromMarkers();
            Invalidate();

            if ( !suppressEvent && GradientChanged != null )
                GradientChanged( this.Parent, EventArgs.Empty );
            mMarkers.Change();
        }

        protected override void OnPaint( PaintEventArgs e )
        {
            base.OnPaintBackground( e );

            Graphics g          = e.Graphics;
            // Determine what to draw. Don't rebuild the gradient when we have a 
            // cahced bitmap, or when there is a marker being moved.
            bool suppressBmp    = ( movingMarker != null );
            
            // Draw the markers
            DrawMarkers( g );

            // Paint the checkerboard style background 
            using ( HatchBrush hb = new HatchBrush( HatchStyle.LargeCheckerBoard, Color.White, Color.LightGray ) )
                g.FillRectangle( hb, gradBarRect );

            // If we are rebuilding the gradient, get the appropriate color blend
            if ( !isBitmapInvalid )
	            g.DrawImage( this.gradientBmp, gradBarRect );
            else if ( suppressBmp )
            {
                // Simply paint the gradient when a marker is being moved. No use
                // painting it on to a bitmap just yet.
                this.gradient.PaintGradientWithDirectionOverride(
                                g, gradBarRect, LinearGradientMode.Horizontal );
            }
            else
            {
                if ( this.gradientBmp == null )
                    this.gradientBmp    = new Bitmap( gradBarRect.Width,
                                                gradBarRect.Height,
                                                PixelFormat.Format32bppArgb );

                // Cache the gradient by painting it onto a bitmap
                Graphics gBmp    = Graphics.FromImage( gradientBmp );
                Rectangle bmpRct = new Rectangle( 0, 0, gradBarRect.Width, gradBarRect.Height );

                this.gradient.PaintGradientWithDirectionOverride(
                                            gBmp, bmpRct,
                                            LinearGradientMode.Horizontal );
                g.DrawImage( this.gradientBmp, gradBarRect );
            }

            // Draw the border around the gradient
            g.DrawRectangle( Pens.Black, gradBarRect );
        }

        private void DrawMarkers(Graphics g)
        {
            foreach (Marker mrk in mMarkers.GetMarkers())
            {
                mrk.UpdateBounds( compansatedMarkerStrip );//using ( SolidBrush br = new SolidBrush( mrk.Color ) )
                if ( mrk.Equals(this.selectedMarker) )
                    g.FillRectangle( Brushes.Black, mrk.Bounds );
                else
                    g.FillRectangle( Brushes.White, mrk.Bounds );
                g.DrawRectangle( Pens.Black,
                            mrk.Bounds.X,
                            mrk.Bounds.Y,
                            mrk.Bounds.Width,
                            mrk.Bounds.Height );
            }
        }

        private ColorBlend GetBlendFromMarkers()
        {
            Color[] colors = new Color[mMarkers.GetMarkers().Count];
            float[] positions = new float[mMarkers.GetMarkers().Count];

            for ( int i = 0; i < colors.Length; i++ )
            {
                colors[i] = mMarkers.GetMarkers()[i].Color;
                positions[i] = mMarkers.GetMarkers()[i].Position;
            }
            ColorBlend blend = new ColorBlend(mMarkers.GetMarkers().Count);
            blend.Colors        = colors;
            blend.Positions     = positions;
            return blend;
        }

        #endregion

        #region Marker Management
        
        protected virtual void OnMarkerSelected( Marker selectedMarker )
        {
            this.selectedMarker = selectedMarker;

            Rectangle invalidRect   = markerStrip;
            invalidRect.Inflate( (int)selectedMarker.Bounds.Width, 0 );
            this.Invalidate( invalidRect );

            if ( MarkerSelected != null )
                MarkerSelected( selectedMarker, EventArgs.Empty );
        }

        #region Marker Movement

        protected override void OnMouseDown( MouseEventArgs e )
        {
            // Can a marker (other than the first or the last) receive this 
            // mouse down event?
            for (int i = 1; i < mMarkers.GetMarkers().Count - 1; i++)
                if (mMarkers.GetMarkers()[i].Bounds.Contains(e.X, e.Y))
                {
                    movingMarker = mMarkers.GetMarkers()[i];
                    return;
                }
        }

        protected override void OnMouseMove( MouseEventArgs e )
        {
            // Move the marker along the marker strip. If the marker is dragged out
            // of the marker strip then the marker is removed.
            if ( markerStrip.Contains( e.X, e.Y ) && movingMarker != null )
            {
                float a = markerStrip.Width;
                movingMarker.SetPositionSilently( e.X / a );
                OnGradientChanged(silentMarkers);
            }
            else if ( movingMarker != null ) // Remove the marker
            {
                this.mMarkers.GetMarkers().Remove(movingMarker);
                this.movingMarker   = null;
                OnGradientChanged(false);
            }
        }

        protected override void OnMouseUp( MouseEventArgs e )
        {
            if ( movingMarker != null ) // A movement was done. Resort the array
            {
                mMarkers.GetMarkers().Sort(CompareMarkers);
                Invalidate( gradBarRect );
                selectedMarker  = movingMarker;

                OnMarkerSelected( movingMarker );
                // Notify that we have a new gradient
                OnGradientChanged(false);
            }
            else if (mMarkers.GetMarkers()[0].Bounds.Contains(e.X, e.Y))
                OnMarkerSelected(mMarkers.GetMarkers()[0]);
            else if (mMarkers.GetMarkers()[mMarkers.GetMarkers().Count - 1].Bounds.Contains(e.X, e.Y))
                OnMarkerSelected(mMarkers.GetMarkers()[mMarkers.GetMarkers().Count - 1]);

            movingMarker    = null;
        }

        private static int CompareMarkers( Marker a, Marker b )
        {
            return a.Position.CompareTo( b.Position );
        }

        #endregion

        void marker_MarkerUpdated( object sender, EventArgs e )
        {
            OnGradientChanged(false);
            
        }

        protected override void OnMouseDoubleClick( MouseEventArgs e )
        {
            // If an location on the marker strip is double clicked where there
            // are no existing markers, a new marker is added.
            Point pt = new Point( e.X, e.Y );

            if ( markerStrip.Contains( pt ) )
            {
                int i = 0;
                for (; i < mMarkers.GetMarkers().Count; i++)
                    if (mMarkers.GetMarkers()[i].Bounds.Contains(pt)) 
                        return;
                    else if (mMarkers.GetMarkers()[i].Bounds.X > pt.X)
                        break;

                float a, b;
                a   = (float)e.X - DEFAULT_PADDING;
                b   = (float)markerStrip.Width;

                Marker newMarker        = new Marker( a/b );
                newMarker.UpdateBounds( compansatedMarkerStrip );

                // Select the color for the new marker. It should be the current 
                // color of the gradient at the marker point.
                if ( this.gradientBmp != null )
                    newMarker.Color = gradientBmp.GetPixel( 
                                            (int)newMarker.Bounds.X, 0 );

                newMarker.MarkerUpdated += new EventHandler( marker_MarkerUpdated );
                mMarkers.GetMarkers().Insert(i, newMarker);
                
                // Notify that we have a new gradient
                OnGradientChanged(false);
                OnMarkerSelected( newMarker );
            }
        }

        #endregion

        #region Construction

        public GradientBuilder()
        {

            mMarkers = new MarkerList();
            SetStyle( ControlStyles.AllPaintingInWmPaint, true );
            SetStyle( ControlStyles.OptimizedDoubleBuffer, true );
            SetStyle( ControlStyles.ResizeRedraw, true );
            SetStyle( ControlStyles.UserPaint, true );

            this.gradient                   = new Gradient();
            this.gradient.GradientDirection = LinearGradientMode.Horizontal;
            this.mMarkers.SetMarkers( new List<Marker>(2));

            Marker startMarker, endMarker;
            startMarker   = new Marker( 0.0f, Color.Black, false );
            endMarker     = new Marker( 1.0f, Color.White, false );
            mMarkers.GetMarkers().Add(startMarker);
            mMarkers.GetMarkers().Add(endMarker);

            startMarker.MarkerUpdated   += new EventHandler( marker_MarkerUpdated );
            endMarker.MarkerUpdated     += new EventHandler( marker_MarkerUpdated );

            
        }

        #endregion
    }
}
