using System;
using System.Diagnostics;


namespace RakabrotPluginInterface
{
    namespace RMath
    {
        public struct Complex
        {
            public double Real;

            public double Imag;

            public Complex(double real, double imag)
            {
                this.Real = real;
                this.Imag = imag;
            }

            public Complex Conjugate
            {
                get { return new Complex(Real, -Imag); }
            }

            public double Norm
            {
                get { return Real * Real + Imag * Imag; }
            }

            public double Modulus
            {
                get { return System.Math.Sqrt(Real * Real + Imag * Imag); }
            }

            public double Argument
            {
                get { return System.Math.Atan2(Imag, Real); }
            }

            public static Complex Polar(double modulus, double argument)
            {
                return new Complex(
                    modulus * System.Math.Cos(argument),
                    modulus * System.Math.Sin(argument));
            }

            public static Complex operator +(Complex z1, Complex z2)
            {
                return new Complex(z1.Real + z2.Real, z1.Imag + z2.Imag);
            }

            public static Complex operator -(Complex z1, Complex z2)
            {
                return new Complex(z1.Real - z2.Real, z1.Imag - z2.Imag);
            }

            public static Complex operator *(Complex z1, Complex z2)
            {
                return new Complex(
                    z1.Real * z2.Real - z1.Imag * z2.Imag,
                    z1.Real * z2.Imag + z1.Imag * z2.Real);
            }

            public static Complex operator /(Complex z1, Complex z2)
            {
                double value = z2.Real * z2.Real + z2.Imag * z2.Imag;

                return new Complex(
                    (z1.Real * z2.Real + z1.Imag * z2.Imag) / value,
                    (z1.Imag * z2.Real - z1.Real * z2.Imag) / value);
            }

            public static bool operator ==(Complex z1, Complex z2)
            {
                return (z1.Real == z2.Real && z1.Imag == z2.Imag);
            }

            public static bool operator !=(Complex z1, Complex z2)
            {
                return (z1.Real != z2.Real || z1.Imag == z2.Imag);
            }

            public static Complex Cos(Complex z)
            {
                Complex z1 = Exp(new Complex(-z.Imag, z.Real));
                Complex z2 = Exp(new Complex(z.Imag, -z.Real));

                return new Complex(0.5 * (z1.Real + z2.Real), 0.5 * (z1.Imag + z2.Imag));
            }

            public static Complex Cosh(Complex z)
            {
                Complex z1 = Exp(z);
                Complex z2 = Exp(new Complex(-z.Real, -z.Imag));

                return new Complex(0.5 * (z1.Real + z2.Real), 0.5 * (z1.Imag + z2.Imag));
            }

            public static Complex Sin(Complex z)
            {
                Complex z1 = Exp(new Complex(-z.Imag, z.Real));
                Complex z2 = Exp(new Complex(z.Imag, -z.Real));

                return new Complex(0.5 * (z1.Imag - z2.Imag), 0.5 * (z2.Real - z1.Real));
            }

            public static Complex Sinh(Complex z)
            {
                Complex z1 = Exp(z);
                Complex z2 = Exp(new Complex(-z.Real, -z.Imag));

                return new Complex(0.5 * (z1.Real - z2.Real), 0.5 * (z1.Imag - z2.Imag));
            }

            public static Complex Tan(Complex z)
            {
                return Sin(z) / Cos(z);
            }

            public static Complex Tanh(Complex z)
            {
                return Sinh(z) / Cosh(z);
            }

            public static Complex Exp(Complex z)
            {
                double nvalue = System.Math.Exp(z.Real);

                return new Complex(
                    nvalue * System.Math.Cos(z.Imag),
                    nvalue * System.Math.Sin(z.Imag));
            }

            public static Complex Log(Complex z)
            {
                return new Complex(System.Math.Log(z.Modulus), z.Argument);
            }

            public static Complex Log10(Complex z)
            {
                const double log10 = 2.3025850929940459;

                Complex value = Log(z);

                value.Real /= log10;
                value.Imag /= log10;

                return value;
            }

            public static Complex Pow(Complex baseNumber, Complex index)
            {
                return Exp(index * Log(baseNumber));
            }
            public Complex PowInt(Complex baseNumber, int index)
            {
                Complex Result = new Complex(1, 0);
                for (int i = 1; i < index; i++)
                {
                    Result = Result * baseNumber;
                }
                return Result;
            }

            public Complex Sqrt(Complex z)
            {
                double value = System.Math.Sqrt(z.Real * z.Real + z.Imag * z.Imag) + z.Real;

                return new Complex(
                    System.Math.Sqrt(0.5 * value),
                    System.Math.Sqrt(0.5 / value) * z.Imag);
            }

            public override string ToString()
            {
                return (String.Format("{0} + {1}i", Real, Imag));
            }

            public override bool Equals(object obj)
            {
                return base.Equals(obj);
            }

            public override int GetHashCode()
            {
                return Real.GetHashCode() ^ Imag.GetHashCode();
            }
            public Complex SetReal(double R)
            {
                Real = R;
                return this;
            }
            public Complex SetImag(double I)
            {
                Imag = I;
                return this;
            }
            public Complex AddReal(double aR)
            {
                Real += aR;
                return this;
            }
            public Complex AddImag(double aI)
            {
                Imag += aI;
                return this;
            }
        }
    }
}